/*
 * SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2023-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: LicenseRef-NvidiaProprietary
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include <config.h>

#include <doca_flow.h>
#include <sys/types.h>

#include "doca-mirror.h"
#include "id-fpool.h"
#include "netdev-doca.h"
#include "netdev-offload.h"
#include "openvswitch/vlog.h"
#include "openvswitch/list.h"
#include "ovs-doca.h"
#include "refmap.h"
#include "util.h"

/*
 * Mirrros in OVS DOCA OFFLOAD:
 *
 * Upper layer (netdev-dpdk-offload) provides ovs doca offload layer (dpdk-offload-doca) an
 * ordered RTE_FLOW_ACTION list. Mirrors are all the output ports before the last forward
 * action (JUMP, OUTPUT, ...).
 * Upper layer put those mirrored outputs in a OVS_RTE_FLOW_ACTION_TYPE(NESTED) which
 * has a nested sample action with ratio 1 that executes a sample action list with an output.
 *
 * Each such nested action output is taken to a shared resource mirror configuration target.
 * The last forward action is taken to the entry's forward.
 *
 * All shared mirror resources are created upon ESW initialization, configured
 * with dummy configuration and bound to the ESW port.
 * Upon allocating a new configuration, such a resource is taken from the pool and re-configured
 * with the required configuration.
 *
 * Example of action list with mirrors:
 * set_ipv4_src ipv4_addr 1.1.1.1 / nested_actions(represented_port ethdev_port_id 3) /
 * nested_actions(represented_port ethdev_port_id 2) / represented_port ethdev_port_id 1 / end
 * In this case, the nested output ports 2 and 3 are taken to a mirror shared resource
 * targets (nr_targets=2) and used as a key to find a similar shared mirror resource.
 * The created/found shared resource is set as the entry's monitor, and output to port 1
 * will be the entry's forward (together with entry's action to set ipv4).
 *
 * Example of action list with mirrors and needing an action split (done by upper layer):
 * nested(out1), nested(out2), push vlan, nested(out3), out4
 *
 * This action list will cause the original rule will be split to two rules:
 *   Rule 1 actions: nested(out1), nested(out2), set_tag, forward to continuation table
 *   Rule 2 actions: push vlan, nested(out3), out4
 *
 * Rule 1 will create a shared mirror with nr_targets=2, targets=1,2
 * Rule 2 will create a shared mirror with nr_targets=1, targets=3
 */

VLOG_DEFINE_THIS_MODULE(doca_mirrors);

struct doca_mirrors {
    uint32_t base_id;
    struct refmap *rfm;
    struct id_fpool *id_pool;
};

struct doca_mirror_ctx {
    struct doca_mirrors *doca_mirrors;
    uint32_t id;
};

OVS_ASSERT_PACKED(struct doca_mirror_key,
    struct doca_flow_mirror_target targets[DOCA_MIRRORS_MAX_TARGETS];
    char pad[4];
    int nr_targets;
);

struct doca_mirror_arg {
    struct doca_mirrors *doca_mirrors;
    struct doca_mirror_key *key;
};

static int
doca_mirror_ctx_init(void *ctx_, void *arg_)
{
    struct doca_flow_shared_resource_cfg resource_cfg;
    unsigned int tid = netdev_offload_thread_id();
    struct doca_mirror_ctx *mirror_ctx = ctx_;
    struct doca_mirror_arg *arg = arg_;
    struct doca_mirror_key *key;
    int err;

    mirror_ctx->doca_mirrors = arg->doca_mirrors;

    if (!id_fpool_new_id(mirror_ctx->doca_mirrors->id_pool, tid, &mirror_ctx->id)) {
        return -1;
    }

    key = refmap_key_from_value(mirror_ctx->doca_mirrors->rfm, mirror_ctx);
    resource_cfg = (struct doca_flow_shared_resource_cfg) {
        .mirror_cfg.nr_targets = key->nr_targets,
        .mirror_cfg.target = key->targets,
    };

    err = doca_flow_shared_resource_set_cfg(DOCA_FLOW_SHARED_RESOURCE_MIRROR,
                                            mirror_ctx->id, &resource_cfg);
    if (err != DOCA_SUCCESS) {
        VLOG_ERR("Failed to set resource_cfg of shared mirror (id %d), err %d - %s",
                 mirror_ctx->id, err, doca_error_get_descr(err));
    }

    return err;
}

static int
doca_mirror_set_cfg_dummy(uint32_t id)
{
    struct doca_flow_shared_resource_cfg dummy_cfg = {
        .mirror_cfg = (struct doca_flow_resource_mirror_cfg) {
            .nr_targets = 1,
            .target = (struct doca_flow_mirror_target[]) {
                [0] = {
                    .fwd.type = DOCA_FLOW_FWD_DROP,
                },
            },
            .fwd = { .type = DOCA_FLOW_FWD_NONE, },
        },
    };

    return doca_flow_shared_resource_set_cfg(DOCA_FLOW_SHARED_RESOURCE_MIRROR,
                                             id, &dummy_cfg);
}

static void
doca_mirror_ctx_uninit(void *ctx_)
{
    unsigned int tid = netdev_offload_thread_id();
    struct doca_mirror_ctx *mirror_ctx = ctx_;

    doca_mirror_set_cfg_dummy(mirror_ctx->id);
    id_fpool_free_id(mirror_ctx->doca_mirrors->id_pool, tid, mirror_ctx->id);
}

static struct ds *
doca_mirror_ctx_dump(struct ds *s, void *key_, void *ctx_ OVS_UNUSED, void *arg_)
{
    struct doca_mirror_key *key = key_;
    struct doca_mirror_arg *arg = arg_;
    int i;

    ds_put_format(s, "doca_mirrors: %p, nr_targets=%d", arg->doca_mirrors, key->nr_targets);
    for (i = 0; i < key->nr_targets; i++) {
        if (key->targets[i].has_encap) {
            ds_put_format(s, ", target[%d].has_encap", i);
        }
        if (key->targets[i].fwd.type == DOCA_FLOW_FWD_PORT) {
            ds_put_format(s, ", target[%d].fwd.port_id=%d", i, key->targets[i].fwd.port_id);
        } else if (key->targets[i].fwd.type == DOCA_FLOW_FWD_PIPE) {
            ds_put_format(s, ", target[%d].fwd.next_pipe=%p", i, key->targets[i].fwd.next_pipe);
        } else {
            ds_put_format(s, ", target[%d].fwd.type=%d", i, key->targets[i].fwd.type);
        }
    }

    return s;
}

struct doca_mirrors *
doca_mirrors_create(uint32_t esw_id, struct netdev *netdev)
{
    struct doca_flow_port *esw_port = netdev_doca_port_get(netdev);
    uint32_t ids[OVS_DOCA_MAX_MIRRORS_PER_ESW];
    struct doca_mirrors *doca_mirrors;
    uint32_t base_id;
    int i, err;

    base_id = esw_id * OVS_DOCA_MAX_MIRRORS_PER_ESW;
    for (i = 0; i < OVS_DOCA_MAX_MIRRORS_PER_ESW; i++) {
        ids[i] = base_id + i;

        err = doca_mirror_set_cfg_dummy(ids[i]);
        if (err != DOCA_SUCCESS) {
            VLOG_ERR("Failed to set resource_cfg of shared mirror (id %d), err %d - %s",
                     ids[i], err, doca_error_get_descr(err));
            return NULL;
        }
    }

    err = doca_flow_shared_resources_bind(DOCA_FLOW_SHARED_RESOURCE_MIRROR,
                                          ids, OVS_DOCA_MAX_MIRRORS_PER_ESW, esw_port);
    if (err) {
        return NULL;
    }

    doca_mirrors = xzalloc(sizeof *doca_mirrors);
    doca_mirrors->id_pool = id_fpool_create(MAX_OFFLOAD_THREAD_NB,
                                            esw_id * OVS_DOCA_MAX_MIRRORS_PER_ESW,
                                            OVS_DOCA_MAX_MIRRORS_PER_ESW);
    if (!doca_mirrors->id_pool) {
        goto err_create_id_pool;
    }

    doca_mirrors->rfm = refmap_create("doca-mirrors", sizeof(struct doca_mirror_key),
                                      sizeof(struct doca_mirror_ctx),
                                      doca_mirror_ctx_init,
                                      doca_mirror_ctx_uninit,
                                      doca_mirror_ctx_dump,
                                      false);
    if (!doca_mirrors->rfm) {
        goto err_create_refmap;
    }

    return doca_mirrors;

err_create_refmap:
    id_fpool_destroy(doca_mirrors->id_pool);
err_create_id_pool:
    free(doca_mirrors);
    return NULL;
}

void
doca_mirrors_destroy(struct doca_mirrors *doca_mirrors)
{
    if (!doca_mirrors) {
        return;
    }

    refmap_destroy(doca_mirrors->rfm);
    id_fpool_destroy(doca_mirrors->id_pool);
    free(doca_mirrors);
}

struct doca_mirror_ctx *
doca_mirror_ctx_ref(struct doca_mirrors *doca_mirrors,
                    struct doca_flow_resource_mirror_cfg *cfg)
{
    struct doca_mirror_key key = {
        .nr_targets = cfg->nr_targets,
    };
    struct doca_mirror_arg arg = {
        .doca_mirrors = doca_mirrors,
        .key = &key,
    };

    if (cfg->nr_targets == 0 || cfg->nr_targets > DOCA_MIRRORS_MAX_TARGETS ||
        cfg->fwd.type != DOCA_FLOW_FWD_NONE) {
        return NULL;
    }

    memcpy(&key.targets, cfg->target, sizeof(*cfg->target) * cfg->nr_targets);
    return refmap_ref(doca_mirrors->rfm, &key, &arg);
}

void
doca_mirror_ctx_unref(struct doca_mirror_ctx *mirror_ctx)
{
    if (!mirror_ctx) {
        return;
    }

    refmap_unref(mirror_ctx->doca_mirrors->rfm, mirror_ctx);
}

int
doca_mirror_ctx_get_id(struct doca_mirror_ctx *mirror_ctx, uint32_t *id)
{
    if (!mirror_ctx || !id) {
        return -1;
    }

    *id = mirror_ctx->id;
    return 0;
}
